"""user.py contains basic information about the steno3d user"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

import properties


class User(properties.PropertyClass):
    """Class representing a user instance"""
    _model_api_location = "user"
    email = properties.String('email')
    name = properties.String('name')
    url = properties.String('url')
    affiliation = properties.String('affiliation')
    location = properties.String('location')
    username = properties.String('username')

    file_size_limit = properties.Int(
        'Inidividual file limit',
        default=5000000
    )
    project_size_limit = properties.Int(
        'Project size limit',
        default=25000000
    )
    project_resource_limit = properties.Int(
        'Maximum resources in a project',
        default=15
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._on_property_change = self._immutable_properties

    def _immutable_properties(self, prop, pre, post):
        if not pre == post:
            print('User data must be modified in settings at steno3d.com')
            setattr(self, '_p_' + prop, pre)



"""base.py sets up BaseExample class that allows examples to be accessed
as properties at the class level
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os import mkdir
from os.path import exists
from os.path import expanduser
from os.path import isdir
from os.path import realpath
from os.path import sep

from future.utils import with_metaclass
from six import string_types
from zipfile import ZipFile

from future.standard_library import install_aliases
install_aliases()
from urllib.request import urlretrieve                          # nopep8



class exampleproperty(object):
    """wrapper that sets class method as saved property"""

    def __init__(self, func):
        self.func = classmethod(func)
        self.propname = '_' + func.__name__

    def __get__(self, cls, owner):
        if getattr(owner, self.propname, None) is None:
            setattr(owner, self.propname, self.func.__get__(None, owner)())
        return getattr(owner, self.propname)


class exampleoutput(object):
    """wrapper that sets class method to remove saved properties"""

    def __init__(self, func, directory):
        self.func = classmethod(func)
        self._directory = ['_' + prop for prop in directory]
        self._stash = [None for prop in directory]

    def __get__(self, cls, owner):
        def _output():
            """return the original function surrounded by stash and unstash"""
            self.stash(owner)
            out = self.func.__get__(None, owner)()
            self.unstash(owner)
            return out
        return _output

    def stash(self, owner):
        """stash saved properties during example output"""
        for i, prop in enumerate(self._directory):
            self._stash[i] = getattr(owner, prop, None)
            setattr(owner, prop, None)

    def unstash(self, owner):
        """unstash saved properties after example output"""
        for i, prop in enumerate(self._directory):
            setattr(owner, prop, self._stash[i])


class _ExampleMetaClass(type):
    """metaclass that wraps every function with exampleproperty()"""

    def __new__(mcs, name, bases, attrs):
        if name not in ('BaseExample', 'Images'):
            assert ('get_resources' in attrs and
                    callable(attrs['get_resources'])), \
                'Example {cls} class must have method get_resources()'.format(
                    cls=name
                )
            assert ('get_project' in attrs and
                    callable(attrs['get_project'])), \
                'Example class {cls} must have method get_project()'.format(
                    cls=name
                )

        example_directory = []
        for attr in attrs:
            value = attrs[attr]
            if not attr.startswith('get_') and not attr == 'fetch_data' and \
                    callable(value):
                attrs[attr] = exampleproperty(value)
                example_directory.append(attr)

        for attr in attrs:
            value = attrs[attr]
            if attr.startswith('get_') and callable(value):
                attrs[attr] = exampleoutput(value, example_directory)

        return type.__new__(mcs, name, bases, attrs)


class BaseExample(with_metaclass(_ExampleMetaClass, object)):
    """basic class that all examples inherit from"""

    def __init__(self, *args, **kwargs):
        raise TypeError('Examples cannot be instantiated. Please access '
                        'the properties directly.')

    def example_name(self):
        return 'BaseExample'

    def data_directory(self):
        """path to directory containing all assets"""
        return sep.join([expanduser('~'), '.steno3d_python_assets'])

    def sub_directory(self):
        return self.example_name.lower()

    def filenames(self):
        return []

    def data_url(self):
        return 'https://storage.googleapis.com/steno3d-examples'

    @classmethod
    def fetch_data(cls, directory=None, download_if_missing=True,
                   filename=None, verbose=True):
        if filename is not None and not isinstance(filename, string_types):
            raise ValueError('filename: must be the name of one file')
        if cls.filenames == []:
            raise ValueError('Example does not require any files')
        if directory is not None and not isinstance(directory, string_types):
            raise ValueError('directory: must be the name of a directory')
        if directory is not None:
            directory = realpath(expanduser(directory))
            if not isdir(directory):
                raise ValueError(
                    '{}: directory does not exist'.format(directory)
                )
            cls._data_directory = directory
        if verbose:
            print('Fetching data...')
        destination = sep.join([cls.data_directory, cls.sub_directory])
        archive = sep.join([destination, cls.sub_directory + '.zip'])
        if download_if_missing and not exists(cls.data_directory):
            mkdir(cls.data_directory)
        if download_if_missing and not exists(destination):
            mkdir(destination)
        if verbose:
            print('Destination: ' + destination)
        filenames = cls.filenames if filename is None else [filename]
        for fname in filenames:
            if verbose:
                print('    Fetching: ' + fname)
            destination_file = sep.join([destination, fname])
            if exists(destination_file):
                if verbose:
                    print('        Local copy found')
                if filename is not None:
                    return destination_file
                continue
            if not exists(archive) and download_if_missing:
                if verbose:
                    print('        Downloading archive...')
                url = '/'.join([cls.data_url, cls.sub_directory + '.zip'])
                urlretrieve(url, archive)
                if verbose:
                    print('        Archive downloaded successfully')
            if exists(archive):
                if verbose:
                    print('        Local archive found: extracting...')
                try:
                    zf = ZipFile(archive)
                    zf.extract(fname, destination)
                    if verbose:
                        print('        File extracted successfully')
                    if filename is not None:
                        return destination_file
                except KeyError:
                    raise IOError('Required file(s) not found in '
                                  'archive. Archive file may be out of '
                                  'date. Please delete archive and '
                                  '{exclass}.fetch_data() again: '
                                  '{archfile}'.format(
                                    exclass=cls.example_name,
                                    archfile=archive
                                  ))
                continue
            raise IOError('Required file(s) not found - please call '
                          '{exclass}.fetch_data() to download and '
                          'save to a default folder in your home '
                          'directory or {exclass}.'
                          'fetch_data(directory=your_local_directory) '
                          'to set an alternative local directory.'.format(
                            exclass=cls.example_name
                          ))



"""surface.py contains the Surface composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from json import dumps

from numpy import max as npmax
from numpy import min as npmin
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh2DOptions(MeshOptions):
    pass


class _SurfaceOptions(ColorOptions):
    pass


class Mesh2D(BaseMesh):
    """class steno3d.Mesh2D

    Contains spatial information about a 2D surface defined by
    triangular faces.
    """
    vertices = properties.Array(
        'Mesh vertices',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    triangles = properties.Array(
        'Mesh triangle vertex indices',
        shape=('*', 3),
        dtype=int,
        required=True
    )
    opts = properties.Pointer(
        'Mesh2D Options',
        ptype=_Mesh2DOptions
    )

    @property
    def nN(self):
        """ get number of nodes """
        return len(self.vertices)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.triangles)

    def nbytes(self, name=None):
        if name in ('vertices', 'triangles'):
            return getattr(self, name).astype('f4').nbytes
        elif name is None:
            return self.nbytes('vertices') + self.nbytes('triangles')
        raise ValueError('Mesh2D cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name in ('vertices', 'triangles'):
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if npmin(self.triangles) < 0:
            raise ValueError('Triangles may only have positive integers')
        if npmax(self.triangles) >= len(self.vertices):
            raise ValueError('Triangles expects more vertices than provided')
        self._validate_file_size('vertices')
        self._validate_file_size('triangles')
        return True


    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'vertices' in dirty or force:
            files['vertices'] = \
                self._properties['vertices'].serialize(self.vertices)
        if 'triangles' in dirty or force:
            files['triangles'] = \
                self._properties['triangles'].serialize(self.triangles)
        return files



class Mesh2DGrid(BaseMesh):
    """Contains spatial information of a 2D grid."""
    h1 = properties.Array(
        'Grid cell widths, x-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h2 = properties.Array(
        'Grid cell widths, y-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    x0 = properties.Vector(
        'Origin vector',
        default=[0, 0, 0]
    )
    Z = properties.Array(
        'Node topography',
        shape=('*',),
        dtype=float
    )
    opts = properties.Pointer(
        'Mesh2D Options',
        ptype=_Mesh2DOptions
    )

    @property
    def nN(self):
        """ get number of nodes """
        return (len(self.h1)+1) * (len(self.h2)+1)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.h1) * len(self.h2)

    def nbytes(self, name=None):
        filenames = ('h1', 'h2', 'x0', 'Z')
        if name in filenames:
            if getattr(self, name, None) is None:
                return 0
            return getattr(self, name).astype('f4').nbytes
        if name is None:
            return sum(self.nbytes(fn) for fn in filenames)
        raise ValueError('Mesh2DGrid cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name in ('h1', 'h2', 'x0', 'Z'):
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.x0.nV != 1:
            raise ValueError('Origin x0 must be only one vector')
        if getattr(self, 'Z', None) is not None and len(self.Z) != self.nN:
            raise ValueError(
                'Length of Z, {zlen}, must equal number of nodes, '
                '{nnode}'.format(
                    zlen=len(self.Z),
                    nnode=self.nN
                )
            )
        self._validate_file_size('h1')
        self._validate_file_size('h2')
        self._validate_file_size('x0')
        self._validate_file_size('Z')
        return True

    def _get_dirty_data(self, force=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty_props
        if ('h1' in dirty or 'h2' in dirty) or force:
            datadict['tensors'] = dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
            ))
        if ('h1' in dirty or 'h2' in dirty or 'x0' in dirty) or force:
            datadict['OUV'] = dumps(dict(
                O=self.x0.tolist()[0],
                U=[self.h1.sum().astype(float), 0, 0],
                V=[0, self.h2.sum().astype(float), 0],
            ))
        return datadict

    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'Z' in dirty or (force and getattr(self, 'Z', None) is not None):
            files['Z'] = self._properties['Z'].serialize(self.Z)
        return files


class _SurfaceBinder(properties.PropertyClass):
    """Contains the data on a 2D surface with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('FACE', 'CELLCENTER'),
            'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Surface(CompositeResource):
    """Contains all the information about a 2D surface"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=[Mesh2D, Mesh2DGrid],
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_SurfaceBinder,
        repeated=True
    )
    textures = properties.Pointer(
        'Textures',
        ptype='Texture2DImage',
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_SurfaceOptions
    )

    def nbytes(self):
        return (self.mesh.nbytes() + sum(d.data.nbytes() for d in self.data) +
                sum(t.nbytes() for t in self.textures))

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'surface.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Surface, self).validate()
        return True


__all__ = ['Surface', 'Mesh2D', 'Mesh2DGrid']



"""wolfpass.py is a project example using resources from Wolf Pass"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from numpy import load as npload

from .base import BaseExample
from ..data import DataArray
from ..line import Line
from ..line import Mesh1D
from ..point import Mesh0D
from ..point import Point
from ..project import Project
from ..surface import Mesh2D
from ..surface import Surface
from ..texture import Texture2DImage
from ..volume import Mesh3DGrid
from ..volume import Volume


class Wolfpass(BaseExample):

    def example_name(self):
        return 'Wolfpass'

    def filenames(self):
        return [
            'AG_gpt.line.npy',
            'AS_ppm.line.npy',
            'AU_gpt.line.npy',
            'CU_pct.line.npy',
            'CU_pct.vol.npy',
            'CU_pct_0.75_1.0_t.cusurf.npy',
            'CU_pct_0.75_1.0_v.cusurf.npy',
            'CU_pct_1.0_1.25_t.cusurf.npy',
            'CU_pct_1.0_1.25_v.cusurf.npy',
            'CU_pct_1.25_1.5_t.cusurf.npy',
            'CU_pct_1.25_1.5_v.cusurf.npy',
            'CU_pct_gt_1.5_t.cusurf.npy',
            'CU_pct_gt_1.5_v.cusurf.npy',
            'CU_pct_lt_0.75_t.cusurf.npy',
            'CU_pct_lt_0.75_v.cusurf.npy',
            'Density.line.npy',
            'MO_ppm.line.npy',
            'Recov.line.npy',
            'S_pct.line.npy',
            'basement_t.lithsurf.npy',
            'basement_v.lithsurf.npy',
            'boreholes_s.line.npy',
            'boreholes_v.line.npy',
            'dacite_data.line.npy',
            'dacite_t.lithsurf.npy',
            'dacite_v.lithsurf.npy',
            'diorite_early_t.lithsurf.npy',
            'diorite_early_v.lithsurf.npy',
            'diorite_late_t.lithsurf.npy',
            'diorite_late_v.lithsurf.npy',
            'dist_to_borehole.lithsurf.npy',
            'dist_to_borehole.vol.npy',
            'drill_loc_v.point.npy',
            'elevation.toposurf.npy',
            'lithology.xsurf.npy',
            'maxdepth.point.npy',
            'ovb_t.lithsurf.npy',
            'ovb_v.lithsurf.npy',
            'section_number.xsurf.npy',
            'topo_t.toposurf.npy',
            'topo_v.toposurf.npy',
            'topography.png',
            'trench.point.npy',
            'vol_h1.vol.npy',
            'vol_h2.vol.npy',
            'vol_h3.vol.npy',
            'vol_x0.vol.npy',
            'xsect_t.xsurf.npy',
            'xsect_v.xsurf.npy',
        ]

    # Borehole drill Points:
    def drill_vertices(self):
        """drill point vertices"""
        return npload(Wolfpass.fetch_data(filename='drill_loc_v.point.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def drill_data(self):
        """drill location raw data

        list of dicts of 'location' and 'data' with 'title' and 'array'
        """
        raw_data = []
        for npyfile in self.filenames:
            if not npyfile.endswith('.point.npy'):
                continue
            if npyfile.endswith('_v.point.npy'):
                continue
            raw_data += [dict(
                location='N',
                data=DataArray(
                    title=npyfile.split('.')[0],
                    array=npload(Wolfpass.fetch_data(filename=npyfile,
                                                     download_if_missing=False,
                                                     verbose=False))
                )
            )]
        return raw_data

    def drill_points(self):
        """Steno3D point resource for borehole drill locations"""
        return Point(
            project=self._dummy_project,
            mesh=Mesh0D(
                vertices=self.drill_vertices
            ),
            data=self.drill_data,
            title='Borehole Drill Locations'
        )

    # Borehole Lines:
    def borehole_vertices(self):
        """borehole line vertices"""
        return npload(Wolfpass.fetch_data(filename='boreholes_v.line.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def borehole_segments(self):
        """borehole segment vertex indices"""
        return npload(Wolfpass.fetch_data(filename='boreholes_s.line.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def borehole_raw_data(self):
        """list of title/array dictionaries for borehole data"""
        raw_data = []
        for npyfile in self.filenames:
            if not npyfile.endswith('.line.npy'):
                continue
            if (npyfile.endswith('_v.line.npy') or
                    npyfile.endswith('_s.line.npy')):
                continue
            raw_data += [dict(
                title=npyfile.split('.')[0],
                array=npload(Wolfpass.fetch_data(filename=npyfile,
                                                 download_if_missing=False,
                                                 verbose=False))
            )]
        return raw_data


    def borehole_data(self):
        """borehole raw data

        list of dicts of 'location' and 'data' with 'title' and 'array'
        """
        data = []
        for raw_data in self.borehole_raw_data:
            data += [dict(
                location='CC',
                data=DataArray(
                    title=raw_data['title'],
                    array=raw_data['array']
                )
            )]
        return data

    def borehole_lines(self):
        """Steno3D line resource for boreholes"""
        return Line(
            project=self._dummy_project,
            mesh=Mesh1D(
                vertices=self.borehole_vertices,
                segments=self.borehole_segments
            ),
            data=self.borehole_data,
            title='Boreholes'
        )

    # CU Percentage Surfaces:
    def cu_prefixes(self):
        """list of prefixes for the different cu pct surfaces"""
        return [fname[:-13] for fname in self.filenames
                if fname.endswith('_v.cusurf.npy')]

    def cu_vertices(self):
        """list of cu pct surface vertices"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_v.cusurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.cu_prefixes]

    def cu_triangles(self):
        """list of cu pct surface triangles"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_t.cusurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.cu_prefixes]

    def cu_surfaces(self):
        """tuple of  Steno3D surface resources for each CU percent range"""
        cu_surfs = tuple()
        for i, prefix in enumerate(self.cu_prefixes):
            cu_surfs += (Surface(
                project=self._dummy_project,
                mesh=Mesh2D(
                    vertices=self.cu_vertices[i],
                    triangles=self.cu_triangles[i]
                ),
                title=prefix
            ),)
        return cu_surfs

    # Lithology Surfaces:
    def lith_prefixes(self):
        """list of prefixes for the different lithology surfaces"""
        return [fname[:-15] for fname in self.filenames
                if fname.endswith('_v.lithsurf.npy')]

    def lith_vertices(self):
        """list of lithology surface vertices"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_v.lithsurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.lith_prefixes]

    def lith_triangles(self):
        """list of lithology surface triangles"""
        return [npload(Wolfpass.fetch_data(filename=prefix + '_t.lithsurf.npy',
                                           download_if_missing=False,
                                           verbose=False))
                for prefix in self.lith_prefixes]

    def lith_diorite_early_data(self):
        """data for early diorite surface"""
        return dict(
            location='N',
            data=DataArray(
                title='dist_to_borehole',
                array=npload(
                    Wolfpass.fetch_data(
                        filename='dist_to_borehole.lithsurf.npy',
                        download_if_missing=False,
                        verbose=False
                    )
                )
            )
        )

    def lith_dacite_vertices(self):
        """vertices for only the dacite surface"""
        for i, pre in enumerate(self.lith_prefixes):
            if pre[-6:] == 'dacite':
                return self.lith_vertices[i]

    def lith_dacite_triangles(self):
        """triangles for only the dacite surface"""
        for i, pre in enumerate(self.lith_prefixes):
            if pre[-6:] == 'dacite':
                return self.lith_triangles[i]

    def lith_surfaces(self):
        """tuple of  Steno3D surface resources for each CU percent range"""
        lith_surfs = tuple()
        for i, prefix in enumerate(self.lith_prefixes):
            if prefix == 'diorite_early':
                lith_data = [self.lith_diorite_early_data]
            else:
                lith_data = []
            lith_surfs += (Surface(
                project=self._dummy_project,
                mesh=Mesh2D(
                    vertices=self.lith_vertices[i],
                    triangles=self.lith_triangles[i]
                ),
                data=lith_data,
                title=prefix
            ),)
        return lith_surfs

    # Topography Surfaces:
    def topo_vertices(self):
        """topography vertices"""
        return npload(Wolfpass.fetch_data(filename='topo_v.toposurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def topo_triangles(self):
        """topography triangles"""
        return npload(Wolfpass.fetch_data(filename='topo_t.toposurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def topo_image(self):
        """surface image"""
        return Wolfpass.fetch_data(filename='topography.png',
                                   download_if_missing=False,
                                   verbose=False)

    def topo_image_orientation(self):
        """surface image O, U, and V"""
        return dict(
            O=[443200, 491750, 0],
            U=[4425, 0, 0],
            V=[0, 3690, 0]
        )

    def topo_texture(self):
        """topography surface image texture"""
        return Texture2DImage(
            O=self.topo_image_orientation['O'],
            U=self.topo_image_orientation['U'],
            V=self.topo_image_orientation['V'],
            image=self.topo_image
        )

    def topo_data(self):
        """elevation data"""
        return dict(
            location='N',
            data=DataArray(
                title='elevation',
                array=npload(Wolfpass.fetch_data(filename='elevation.toposurf.npy',
                                                 download_if_missing=False,
                                                 verbose=False))
            )
        )

    def topo_surface(self):
        """Steno3D surface resource with topo data and surface imagery"""
        return Surface(
            project=self._dummy_project,
            mesh=Mesh2D(
                vertices=self.topo_vertices,
                triangles=self.topo_triangles
            ),
            data=self.topo_data,
            textures=self.topo_texture,
            title='Topography Surface'
        )

    # Cross-section Surface:
    def xsect_vertices(self):
        """cross section vertices"""
        return npload(Wolfpass.fetch_data(filename='xsect_v.xsurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def xsect_triangles(self):
        """cross section triangles"""
        return npload(Wolfpass.fetch_data(filename='xsect_t.xsurf.npy',
                                          download_if_missing=False,
                                          verbose=False))

    def xsect_data(self):
        """cross section raw data"""
        raw_data = []
        for npyfile in self.filenames:
            if not npyfile.endswith('.xsurf.npy'):
                continue
            if (npyfile.endswith('_v.xsurf.npy') or
                    npyfile.endswith('_t.xsurf.npy')):
                continue
            raw_data += [dict(
                location='CC',
                data=DataArray(
                    title=npyfile.split('.')[0],
                    array=npload(Wolfpass.fetch_data(filename=npyfile,
                                                     download_if_missing=False,
                                                     verbose=False))
                    )
            )]
        return raw_data

    def xsect_surface(self):
        """Steno3D surface resource of cross sections"""
        return Surface(
            project=self._dummy_project,
            mesh=Mesh2D(
                vertices=self.xsect_vertices,
                triangles=self.xsect_triangles
            ),
            data=self.xsect_data,
            title='Cross-Sections'
        )

    # Lithology Volume:
    def lith_tensor(self):
        """(h1, h2, h3, x0) for lith volume"""
        return (
            npload(Wolfpass.fetch_data(filename='vol_h1.vol.npy',
                                       download_if_missing=False,
                                       verbose=False)),
            npload(Wolfpass.fetch_data(filename='vol_h2.vol.npy',
                                       download_if_missing=False,
                                       verbose=False)),
            npload(Wolfpass.fetch_data(filename='vol_h3.vol.npy',
                                       download_if_missing=False,
                                       verbose=False)),
            npload(Wolfpass.fetch_data(filename='vol_x0.vol.npy',
                                       download_if_missing=False,
                                       verbose=False))
        )

    def lith_data(self):
        """raw data for lith volume"""
        raw_data = []
        for npyfile in self.filenames:
            if not npyfile.endswith('.vol.npy'):
                continue
            if npyfile.startswith('vol_'):
                continue
            raw_data += [dict(
                location='CC',
                data=DataArray(
                    title=npyfile.split('.')[0],
                    array=npload(Wolfpass.fetch_data(filename=npyfile,
                                                     download_if_missing=False,
                                                     verbose=False))
                )
            )]
        return raw_data

    def lith_volume(self):
        """Steno3D volume resource of lithology"""
        return Volume(
            project=self._dummy_project,
            mesh=Mesh3DGrid(
                h1=self.lith_tensor[0],
                h2=self.lith_tensor[1],
                h3=self.lith_tensor[2],
                x0=self.lith_tensor[3]
            ),
            data=self.lith_data,
            title='Lithology Volume'
        )

    def project(self):
        """empty Steno3D project"""
        return Project(
            title='Wolf Pass'
        )

    def _dummy_project(self):
        """Steno3D project for initializing resources"""
        return Project()

    def get_resources(self):
        """get a copy of the Wolfpass resources"""
        res = (self.drill_points, self.borehole_lines,)
        res += self.cu_surfaces
        res += self.lith_surfaces
        res += (self.topo_surface, self.xsect_surface, self.lith_volume)
        return res

    def get_project(self):
        """get a copy of the Wolfpass project"""
        proj = self.project
        proj.resources = self.get_resources()
        for r in proj.resources:
            r.project = proj
        return proj

    def get_project_topo(self):
        res = (self.drill_points, self.topo_surface,)
        proj = self.project
        proj.resources = res
        proj.title = 'Topography'
        proj.description = ('Topography, surface imagery, '
                            'and borehole drill locations')
        for r in proj.resources:
            r.project = proj
        return proj

    def get_project_dacite(self):
        res = (self.borehole_lines,)
        res += tuple([r for r in self.lith_surfaces
                      if r.title == 'ovb' or r.title == 'dacite'])
        proj = self.project
        proj.resources = res
        proj.title = 'Wolf Pass'
        proj.description = 'Boreholes and dacite formation'
        for r in proj.resources:
            r.project = proj
        return proj



"""data.py contains resource data structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

import properties

from .base import BaseData


class DataArray(BaseData):
    """Data array with unique values at every point in the mesh"""
    _resource_class = 'array'
    array = properties.Array(
        'Data, unique values at every point in the mesh',
        shape=('*',),
        dtype=(float, int),
        required=True
    )

    def __init__(self, array=None, **kwargs):
        super().__init__(**kwargs)
        if array is not None:
            self.array = array

    def nbytes(self, name=None):
        if name is None or name == 'array':
            return self.array.astype('f4').nbytes
        raise ValueError('DataArray cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name == 'array':
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)


    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        self._validate_file_size('array')
        return True

    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'array' in dirty or force:
            files['array'] = self._properties['array'].serialize(self.array)
        return files

__all__ = ['DataArray']



"""mics.py contains miscellaneous files and components to supplement
Steno3D examples
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .base import BaseExample


class Images(BaseExample):
    """Class containing miscellaneous image files."""

    def example_name(self):
        return 'Images'

    def sub_directory(self):
        return 'basic'

    def filenames(self):
        return ['metal.png',
                'woodplanks.png',
                'steno3d.png',
                'steno3d_logo_text.png',
                'licenses.txt']

    def metal(self):
        """metal texture png"""
        return Images.fetch_data(filename='metal.png',
                                 download_if_missing=False,
                                 verbose=False)

    def wood(self):
        """wood texture png"""
        return Images.fetch_data(filename='woodplanks.png',
                                 download_if_missing=False,
                                 verbose=False)

    def logo(self):
        """wood texture png"""
        return Images.fetch_data(filename='steno3d.png',
                                 download_if_missing=False,
                                 verbose=False)

    def logo_with_text(self):
        """wood texture png"""
        return Images.fetch_data(filename='steno3d_logo_text.png',
                                 download_if_missing=False,
                                 verbose=False)

try:
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



"""client.py contains the functionality to link the python steno3d
client with the steno3d website
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import str
from builtins import input
from functools import wraps
from time import sleep

import keyring
import requests
from six import string_types
from six.moves.urllib.parse import urlparse

from .user import User


__version__ = '0.1.4'

PRODUCTION_BASE_URL = 'https://steno3d.com/'
API_SUBPATH = 'api/'
SLEEP_TIME = .75

WELCOME_HEADER = """

Welcome to the Python client library for Steno3D!

"""

DEVKEY_PROMPT = "If you have a Steno3D developer key, please enter it here > "

WELCOME_MESSAGE = """

If you do not have a Steno3D developer key, you need to request
one from the Steno3D website in order to access the API. Please
log in to the application (if necessary) and request a new key.

{base_url}settings/developer

If you are not yet signed up, you can do that here:

{base_url}signup

When you are ready, please enter the key above, or reproduce this
prompt by calling steno3d.login().

"""

LOGIN_FAILED = """

Oh no! We could not log you in.

The API developer key that you provided could not be validated. You could:

1) Clear your keychain with `steno3d.logout()` and try again
2) Restart your Python kernel and try again
3) Check that you have the correct API key
4) Update steno3d with `pip install --upgrade steno3d`
5) Ask for <help@steno3d.com>
6) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

NOT_CONNECTED = """

Oh no! We could not connect to the Steno3D server.

Please ensure that you are:

1) connected to the Internet!
2) Can connect to Steno3D on  https://steno3d.com

If the issue persists please:

1) Ask for <help@steno3d.com>
2) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

BAD_API_KEY = """

The API developer key should be your username followed by '//' then 36
characters. If you have not requested an API key or if you have lost
your API key, please request a new one at:

{base_url}settings/developer

"""

INVALID_VERSION = """

Your version of steno3d is out of date.

{your_version}
{current_version}

Please update steno3d with `pip install --upgrade steno3d`.

"""



class _Comms(object):
    """Comms controls the interaction between the python client and the
    Steno3D website.
    """

    def __init__(self):
        self._user = None
        self._me = None
        self._base_url = PRODUCTION_BASE_URL

    @property
    def host(self):
        """hostname of url"""
        parseresult = urlparse(self.base_url)
        return parseresult.hostname

    @property
    def url(self):
        """url endpoint for uploading"""
        return self.base_url + API_SUBPATH

    @property
    def base_url(self):
        """base url endpoint for uploading"""
        return getattr(self, '_base_url', PRODUCTION_BASE_URL)

    @base_url.setter
    def base_url(self, value):
        assert isinstance(value, string_types), \
            'Endpoint path must be a string'
        # Patch '/' onto bare URL endpoints
        if not value[-1] == '/':
            value += '/'
        # Check for HTTPS
        parsed = urlparse(value)
        if '.com' in parsed.hostname and parsed.scheme != 'https':
            raise Exception('Live endpoints require HTTPS.')

        self._base_url = value

    @property
    def devel_key(self):
        """developer key acquired from steno3d.com"""
        if getattr(self, '_hard_devel_key', None) is not None:
            return self._hard_devel_key

        key = keyring.get_password('steno3d', self.host)
        if key in {None, 'None'}:
            return None
        return str(key)

    @devel_key.setter
    def devel_key(self, value):
        if value is None:
            del self.devel_key
        else:
            keyring.set_password('steno3d', self.host, value)

    @devel_key.deleter
    def devel_key(self):
        self._hard_devel_key = None
        try:
            keyring.delete_password('steno3d', self.host)
        except (keyring.errors.PasswordDeleteError, RuntimeError):
            # Happens when the key or keychain does not exist
            pass

    def get_user(self):
        if getattr(self, '_me', None) is not None:
            return self._me
        elif getattr(self, '_user', None) is None:
            return None
        else:
            username = self._user['uid']
            email = self._user['email']
            name = self._user['name']
            url = self._user['url']
            affiliation = self._user['affiliation']
            location = self._user['location']
            self._me = User(
                username=username if username is not None else 'None',
                email=email if email is not None else 'None',
                name=name if name is not None else 'None',
                url=url if url is not None else 'None',
                affiliation=affiliation if affiliation is not None else 'None',
                location=location if location is not None else 'None',
            )
            return self._me

    def login(self, devel_key=None, skip_keychain=False, endpoint=None):
        """Login to steno3d.com to allow uploading resources

        Optional arguments:
            devel_key     - API key from steno3d.com. Prompt will appear if
                            this is not provided or saved on the keychain
            skip_keychain - Prevents loading or saving API key on the
                            keychain
            endpoint      - target site, default is steno3d.com
        """
        try:
            keyring.get_password('steno3d', self.host)
        except RuntimeError:
            print('Unable to access keychain. Proceeding to login with '
                  '`skip_keychain=True`. That means you will need to '
                  'reenter your developer API key every time you '
                  'restart the kernel.')
            self.login(devel_key, True, endpoint)
            return
        if endpoint is not None:
            self.base_url = str(endpoint)
        # Check client version first.
        try:
            resp = requests.post(
                self.url + 'client/steno3dpy',
                dict(version=__version__)
            )
        except requests.ConnectionError:
            raise Exception(NOT_CONNECTED)
        if resp.status_code == 200:
            resp_json = resp.json()
            your_ver = resp_json['your_version']
            curr_ver = resp_json['current_version']
            if resp_json['valid'] or curr_ver == '0.0.0':
                pass
            elif (your_ver.split('.')[0] == curr_ver.split('.')[0] and
                  your_ver.split('.')[1] == curr_ver.split('.')[1]):
                print(INVALID_VERSION.format(
                    your_version='Your version: ' + your_ver,
                    current_version='Current version: ' + curr_ver
                ))
            else:
                raise Exception(INVALID_VERSION.format(
                    your_version='Your version: ' + your_ver,
                    current_version='Required version: ' + curr_ver
                ))
        elif resp.status_code == 400:
            resp_json = resp.json()
            print(INVALID_VERSION.format(
                your_version='Your version: ' + __version__,
                current_version='Error: ' + resp_json['reason']
            ))
        # Set devel key next
        if devel_key is not None:
            if skip_keychain:
                self._hard_devel_key = devel_key
            else:
                self.devel_key = devel_key
        if ((skip_keychain and self._hard_devel_key is None) or
                self.devel_key is None):
            print(WELCOME_MESSAGE.format(base_url=self.base_url))
            try:
                devel_key = raw_input(WELCOME_HEADER + DEVKEY_PROMPT)
            except NameError:
                devel_key = input(WELCOME_HEADER + DEVKEY_PROMPT)
            split_key = devel_key.split('//')
            if len(split_key) is not 2 or len(split_key[1]) is not 36:
                self.devel_key = None
                raise Exception(BAD_API_KEY.format(base_url=self.base_url))
            if skip_keychain:
                self._hard_devel_key = devel_key
            else:
                self.devel_key = devel_key
        # Check user
        if getattr(self, '_user', None) is None:
            try:
                resp = requests.get(
                    self.url + 'me',
                    headers={'sshKey': self.devel_key}
                )
            except requests.ConnectionError:
                raise Exception(NOT_CONNECTED)
            if resp.status_code is not 200:
                self.devel_key = None
                raise Exception(LOGIN_FAILED)
            self._user = resp.json()
        # Success
        print(
            'Welcome to Steno3D! You are logged in as @{name}'.format(
                name=self.get_user().username
            )
        )

    def logout(self):
        """Logout current user and remove API key from keyring"""
        self.devel_key = None
        self._me = None
        self._user = None


Comms = _Comms()


def needs_login(func):
    """Wrapper used around functions that need you to be logged in"""
    @wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if Comms.get_user() is None:
            print("Please login: 'steno3d.login()'")
        else:
            return func(self, *args, **kwargs)
    return func_wrapper


def pause():
    """Brief pause on localhost to simulate network delay"""
    if 'localhost' in Comms.url:
        sleep(SLEEP_TIME)


@needs_login
def post(url, data=None, files=None):
    """Post data and files to the steno3d online endpoint"""
    return upload(requests.post, url, data, files)


@needs_login
def put(url, data=None, files=None):
    """Put data and files to the steno3d online endpoint"""
    return upload(requests.put, url, data, files)


@needs_login
def upload(request_fcn, url, data, files):
    """Post data and files to the steno3d online endpoint"""
    data = {} if data is None else data
    files = {} if files is None else files
    filedict = {}
    for filename in files:
        if hasattr(files[filename], 'dtype'):
            filedict[filename] = files[filename].file
            filedict[filename + 'Type'] = files[filename].dtype
        else:
            filedict[filename] = files[filename]
    req = request_fcn(
        Comms.url + url,
        data=data,
        files=filedict,
        headers={'sshKey': Comms.devel_key}
    )
    for key in files:
        files[key].file.close()
    return req


def plot(url):
    """Return an IFrame plot"""
    from IPython.display import IFrame
    return IFrame(url, width='100%', height=500)



"""content.py contains the base class for everything users can create
in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from pprint import pformat

import properties

from .client import pause
from .client import post
from .client import put


class UserContent(properties.PropertyClass):
    """Base class for everything user creates and uploads in steno3d"""
    title = properties.String(
        'Title of the model.',
        default=''
    )
    description = properties.String(
        'Description of the model.',
        default=''
    )
    _sync = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._upload_data = None

    @properties.classproperty
    @classmethod
    def _resource_class(cls):
        """name of the class of resource"""
        if getattr(cls, '__resource_class', None) is None:
            cls.__resource_class = cls.__name__.lower()
        return cls.__resource_class

    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/{className}'.format(
                className=cls._resource_class
            )
        return cls.__model_api_location

    def _upload(self, sync=False, verbose=True, tab_level=''):
        if getattr(self, '_uploading', False):
            return
        try:
            if verbose:
                print(
                    tab_level + ('Uploading '
                                 if getattr(self, '_upload_data', None) is None
                                 else 'Updating ') +
                    self._resource_class + ': ' + self.title
                )
            self._uploading = True
            pause()
            assert self.validate()
            self._upload_dirty(sync, verbose, tab_level + '    ')
            if getattr(self, '_upload_data', None) is None:
                self._post(
                    self._get_dirty_data(force=True),
                    self._get_dirty_files(force=True)
                )
            else:
                dirty_data = self._get_dirty_data()
                dirty_files = self._get_dirty_files()
                if len(dirty_data) > 0 or len(dirty_files) > 0:
                    self._put(dirty_data, dirty_files)
            self._mark_clean(recurse=False)
            self._sync = sync
            if verbose:
                print(tab_level + '... Complete!')
        except Exception as err:
            if self._sync:
                print('Upload failed, turning off syncing. To restart '
                      'syncing, upload() again.')
                self._sync = False
            else:
                raise err
        finally:
            self._uploading = False

    def _get_dirty_data(self, force=False):
        dirty = self._dirty_props
        datadict = dict()
        if 'title' in dirty or force:
            datadict['title'] = self.title
        if 'description' in dirty or force:
            datadict['description'] = self.description
        return datadict

    def _get_dirty_files(self, force=False):
        return {}

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        pass

    def _on_property_change(self, name, pre, post):
        if getattr(self, '_sync', False):
            self._upload(self._sync)

    def _post(self, datadict=None, files=None):
        self._client_upload(post, self._model_api_location,
                            datadict, files)

    def _put(self, datadict=None, files=None):
        pause()
        api_uid = '{mapi}/{uid}'.format(mapi=self._model_api_location,
                                        uid=self._upload_data['uid'])
        self._client_upload(put, api_uid, datadict, files)

    def _client_upload(self, request_fcn, url,
                       datadict=None, files=None):
        req = request_fcn(
            url,
            data=datadict if datadict else tuple(),
            files=files if files else tuple(),
        )
        if isinstance(req, list):
            for rq in req:
                if rq.status_code != 200:
                    try:
                        resp = pformat(rq.json())
                    except ValueError:
                        resp = rq

                    raise Exception(
                        'Upload failed: {location}'.format(
                            location=url,
                        ) +
                        '\ndata: {datadict}\nfiles: {filedict}'.format(
                            datadict=pformat(datadict),
                            filedict=pformat(files),
                        ) +
                        '\nresponse: {response}'.format(
                            response=resp,
                        )
                    )
            self._upload_data = [rq.json() for rq in req]
        else:
            if req.status_code != 200:
                try:
                    resp = pformat(req.json())
                except ValueError:
                    resp = req
                raise Exception(
                    'Upload failed: {location}'.format(
                        location=url,
                    ) +
                    '\ndata: {datadict}\nfiles: {filedict}'.format(
                        datadict=pformat(datadict),
                        filedict=pformat(files),
                    ) +
                    '\nresponse: {response}'.format(
                        response=resp,
                    )
                )
            self._upload_data = req.json()

    @property
    def _json(self):
        """Return a JSON representation of the object"""
        json = getattr(self, '_upload_data', None)
        if json is None:
            raise ValueError('JSON not available: Data not uploaded')
        return json

    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        return True




"""volume.py contains the Volume composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from json import dumps

import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh3DOptions(MeshOptions):
    pass


class _VolumeOptions(ColorOptions):
    pass


class Mesh3DGrid(BaseMesh):
    """Contains spatial information of a 3D grid volume."""

    h1 = properties.Array(
        'Tensor cell widths, x-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h2 = properties.Array(
        'Tensor cell widths, y-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h3 = properties.Array(
        'Tensor cell widths, z-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    x0 = properties.Vector(
        'Origin vector',
        default=[0, 0, 0]
    )
    opts = properties.Pointer(
        'Mesh3D Options',
        ptype=_Mesh3DOptions
    )

    @property
    def nN(self):
        """ get number of nodes """
        return (len(self.h1)+1) * (len(self.h2)+1) * (len(self.h3)+1)

    @property
    def nC(self):
        """ get number of cells """
        return len(self.h1) * len(self.h2) * len(self.h3)

    def nbytes(self, name=None):
        filenames = ('h1', 'h2', 'h3', 'x0')
        if name in filenames:
            return getattr(self, name).astype('f4').nbytes
        elif name is None:
            return sum(self.nbytes(fn) for fn in filenames)
        raise ValueError('Mesh3DGrid cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name in ('h1', 'h2', 'h3', 'x0'):
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.x0.nV != 1:
            raise ValueError('Origin x0 must be only one vector')
        self._validate_file_size('h1')
        self._validate_file_size('h2')
        self._validate_file_size('h3')
        self._validate_file_size('x0')
        return True

    def _get_dirty_data(self, force=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty_props
        # datadict = dict()
        if force or ('h1' in dirty or 'h2' in dirty or 'h3' in dirty):
            datadict['tensors'] = dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
                h3=self.h3.tolist()
            ))
        if force or ('h1' in dirty or 'h2' in dirty or 'h3' in dirty or
                     'x0' in dirty):
            datadict['OUVZ'] = dumps(dict(
                O=self.x0.tolist()[0],
                U=[self.h1.sum().astype(float), 0, 0],
                V=[0, self.h2.sum().astype(float), 0],
                Z=[0, 0, self.h3.sum().astype(float)]
            ))
        return datadict


class _VolumeBinder(properties.PropertyClass):
    """Contains the data on a 3D volume with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('CELLCENTER'),
            # 'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Volume(CompositeResource):
    """Contains all the information about a 3D volume"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh3DGrid,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_VolumeBinder,
        repeated=True,
        required=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_VolumeOptions
    )

    def nbytes(self):
        return self.mesh.nbytes() + sum(d.data.nbytes() for d in self.data)

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location == 'CC'  # in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'volume.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Volume, self).validate()
        return True


__all__ = ['Volume', 'Mesh3DGrid']



"""project.py contains the project class that contains resources
in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super

import properties

from .client import needs_login
from .client import Comms
from .content import UserContent


class Project(UserContent):
    """Steno3D top-level project"""
    _model_api_location = 'project/steno3d'

    resources = properties.Pointer(
        'Project Resources',
        ptype='CompositeResource',
        repeated=True
    )

    public = properties.Bool(
        'Public visibility of project',
        default=False
    )

    _public_online = None

    @classmethod
    def _url_view_from_uid(cls, uid):
        """Get full url from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=Comms.base_url,
            mapi='app',
            uid=uid)
        return url

    def nbytes(self):
        return sum(r.nbytes() for r in self.resources)

    @needs_login
    def upload(self, sync=False, verbose=True, print_url=True):
        """Upload the project"""
        if getattr(self, '_upload_data', None) is None:
            assert self.validate()
            if verbose:
                print('Initializing ' +
                      ('public ' if self.public else 'private ') +
                      'project: ' + self.title)
                if self.public:
                    print('This project will be viewable by everyone.')
            self._post(self._get_dirty_data(force=True, initialize=True))
            self._public_online = self.public
        elif verbose and self._public_online:
            print('This project is PUBLIC. It is viewable by everyone.')
        if verbose and not self._public_online == self.public:
            print('Local privacy changes cannot be applied to '
                  'projects that are already uploaded. To make '
                  'these changes, please use the dashboard on '
                  'steno3d.com.')
        self._upload(sync, verbose)
        if print_url:
            print(self._url)
        return self._url

    @properties.validator
    def validate(self):
        """Check if project resource pointers are correct"""
        for res in self.resources:
            if self not in res.project:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that resources point to containing '
                                 'project.')
        self._validate_project_size()
        return True

    def _validate_project_size(self):
        if Comms.get_user() is not None:
            res_limit = Comms.get_user().project_resource_limit
            if len(self.resources) > res_limit:
                raise ValueError(
                    'Total number of resources in project ({res}) '
                    'exceeds limit: {lim}'.format(res=len(self.resources),
                                                  lim=res_limit)
                )
            size_limit = Comms.get_user().project_size_limit
            if self.nbytes() > size_limit:
                raise ValueError(
                    'Total project size ({file} bytes) exceeds limit: '
                    '{lim} bytes'.format(file=self.nbytes(),
                                         lim=size_limit)
                )
        return True

    def _on_property_change(self, name, pre, post):
        if name == 'resources':
            if pre is None:
                pre = []
            if post is None:
                post = []
            for res in post:
                if res not in pre and self not in res.project:
                    res.project += [self]
            for res in pre:
                if res not in post and self in res.project:
                    res.project = [p for p in res.project
                                   if p is not self]
            if len(set(post)) != len(post):
                post_post = []
                for r in post:
                    if r not in post_post:
                        post_post += [r]
                self.resources = post_post
        super()._on_property_change(name, pre, post)

    def _upload_dirty(self, sync=False, verbose=True, tab_level=''):
        dirty = self._dirty
        if 'resources' in dirty:
            [r._upload(sync, verbose, tab_level) for r in self.resources]

    def _get_dirty_data(self, force=False, initialize=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty_props
        if 'public' in dirty or force:
            datadict['public'] = self.public
        if ('resources' in dirty or force) and not initialize:
            datadict['resourceUids'] = ','.join(
                (r._json['longUid'] for r in self.resources)
            )
        return datadict

    @property
    def _url(self):
        if getattr(self, '_upload_data', None) is not None:
            return self._url_view_from_uid(self._upload_data['uid'])

    @property
    @needs_login
    def url(self):
        """steno3d.com url of project if uploaded"""
        if getattr(self, '_upload_data', None) is None:
            raise Exception('Project not uploaded: Please upload() '
                            'before accessing the URL.')
        return self._url


__all__ = ['Project']



"""texture.py contains the texture resource structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from builtins import super
from collections import namedtuple
from io import BytesIO
from json import dumps

import properties

from .base import BaseTexture2D


FileProp = namedtuple('FileProp', ['file', 'dtype'])


class Texture2DImage(BaseTexture2D):
    """Contains an image that can be mapped to a 2D surface"""

    _resource_class = 'image'

    O = properties.Vector(
        'Origin of the texture',
        required=True
    )
    U = properties.Vector(
        'U axis of the texture',
        required=True
    )
    V = properties.Vector(
        'V axis of the texture',
        required=True
    )
    image = properties.Image(
        'Image file',
        required=True
    )

    def nbytes(self, name=None):
        if name is None or name == 'image':
            self.image.seek(0)
            return len(self.image.read())
        raise ValueError('Texture2DImage cannot calculate the number of '
                         'bytes of {}'.format(name))

    def _on_property_change(self, name, pre, post):
        try:
            if name == 'image':
                self._validate_file_size(name)
        except ValueError as err:
            setattr(self, '_p_' + name, pre)
            raise err
        super()._on_property_change(name, pre, post)

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.O.nV != 1 or self.U.nV != 1 or self.V.nV != 1:
            raise ValueError('O, U, and V must each be only one vector')
        self._validate_file_size('image')
        return True

    def _get_dirty_files(self, force=False):
        dirty = self._dirty_props
        files = dict()
        if 'image' in dirty or force:
            self.image.seek(0)
            copy = BytesIO()
            copy.name = 'texture_copy.png'
            copy.write(self.image.read())
            copy.seek(0)
            files['image'] = FileProp(copy, 'png')
        return files

    def _get_dirty_data(self, force=False):
        datadict = super()._get_dirty_data(force)
        dirty = self._dirty_props
        # datadict = dict()
        if ('O' in dirty or 'U' in dirty or 'V' in dirty) or force:
            datadict['OUV'] = dumps(dict(
                O=[self.O[0][0], self.O[0][1], self.O[0][2]],
                U=[self.U[0][0], self.U[0][1], self.U[0][2]],
                V=[self.V[0][0], self.V[0][1], self.V[0][2]],
            ))
        return datadict

    def _repr_png_(self):
        """For IPython display"""
        if self.image is None:
            return None
        self.image.seek(0)
        return self.image.read()


__all__ = ['Texture2DImage']



"""teapot.py provides an example Steno3D project of a teapot"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import loads

from numpy import array

from .base import BaseExample
from ..point import Mesh0D
from ..point import Point
from ..project import Project
from ..surface import Mesh2D
from ..surface import Surface


class Teapot(BaseExample):
    """Class containing components of teapot project. Components can be
    viewed individually or copied into new resources or projects with
    get_resources() and get_project(), respectively.
    """

    def example_name(self):
        return 'Teapot'

    def filenames(self):
        """teapot json file"""
        return ['teapot.json']

    def data(self):
        """teapot data read from json"""
        json_file = Teapot.fetch_data(filename='teapot.json',
                                      download_if_missing=False,
                                      verbose=False)
        with open(json_file, 'r') as f:
            file_data = loads(f.read())
        return file_data

    def vertices(self):
        """n x 3 numpy array of teapot vertices"""
        return array(self.data['vertices'])

    def triangles(self):
        """n x 3 numpy array of teapot triangle vertex indices"""
        return array(self.data['triangles'])

    def points(self):
        """Steno3D points at teapot vertices"""
        return Point(
            project=self._dummy_project,
            mesh=Mesh0D(
                vertices=self.vertices
            ),
            title='Teapot Vertex Points'
        )

    def surface(self):
        """Steno3D teapot surface"""
        return Surface(
            project=self._dummy_project,
            mesh=Mesh2D(
                vertices=self.vertices,
                triangles=self.triangles
            ),
            title='Teapot Surface'
        )

    def project(self):
        """empty Steno3D project"""
        return Project(
            title='Teapot',
            description='Project with surface and points at vertices'
        )

    def _dummy_project(self):
        """Steno3D project for initializing resources"""
        return Project()

    def get_resources(self):
        """get a copy of teapot resources.

        tuple(teapot surface, teapot points,)
        """
        return (self.surface, self.points,)

    def get_project(self):
        """get a copy of teapot project."""
        proj = self.project
        proj.resources = self.get_resources()
        for r in proj.resources:
            r.project = proj
        return proj
